/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.seatunnel.connectors.seatunnel.cdc.oracle.source.parser;

import io.debezium.ddl.parser.oracle.generated.PlSqlParser;
import io.debezium.relational.Column;
import io.debezium.relational.ColumnEditor;
import lombok.Getter;
import lombok.Setter;
import oracle.jdbc.OracleTypes;

import java.sql.Types;

@Getter
@Setter
public class CustomColumnDefinitionParserListener extends BaseParserListener {
    private ColumnEditor columnEditor;

    public CustomColumnDefinitionParserListener() {}

    @Override
    public void enterColumn_definition(PlSqlParser.Column_definitionContext ctx) {
        if (columnEditor != null) {
            resolveColumnDataType(ctx);
            if (ctx.DEFAULT() != null) {
                this.columnEditor.defaultValueExpression(ctx.column_default_value().getText());
            }
        }
        super.enterColumn_definition(ctx);
    }

    @Override
    public void enterModify_col_properties(PlSqlParser.Modify_col_propertiesContext ctx) {
        if (columnEditor != null) {
            resolveColumnDataType(ctx);
            if (ctx.DEFAULT() != null) {
                columnEditor.defaultValueExpression(ctx.column_default_value().getText());
            }
        }
        super.enterModify_col_properties(ctx);
    }

    // todo use dataTypeResolver instead
    private void resolveColumnDataType(PlSqlParser.Column_definitionContext ctx) {
        columnEditor.name(getColumnName(ctx.column_name()));

        boolean hasNotNullConstraint =
                ctx.inline_constraint().stream().anyMatch(c -> c.NOT() != null);
        columnEditor.optional(!hasNotNullConstraint);

        if (ctx.datatype() == null) {
            if (ctx.type_name() != null
                    && "MDSYS.SDO_GEOMETRY"
                            .equalsIgnoreCase(ctx.type_name().getText().replace("\"", ""))) {
                columnEditor.jdbcType(Types.STRUCT).type("MDSYS.SDO_GEOMETRY");
            }
        } else {
            resolveColumnDataType(ctx.datatype());
        }
    }

    private void resolveColumnDataType(PlSqlParser.Modify_col_propertiesContext ctx) {
        columnEditor.name(getColumnName(ctx.column_name()));

        resolveColumnDataType(ctx.datatype());

        boolean hasNullConstraint =
                ctx.inline_constraint().stream().anyMatch(c -> c.NULL_() != null);
        boolean hasNotNullConstraint =
                ctx.inline_constraint().stream().anyMatch(c -> c.NOT() != null);
        if (hasNotNullConstraint && columnEditor.isOptional()) {
            columnEditor.optional(false);
        } else if (hasNullConstraint && !columnEditor.isOptional()) {
            columnEditor.optional(true);
        }
    }

    private void resolveColumnDataType(PlSqlParser.DatatypeContext ctx) {
        // If the context is null, there is nothing this method can resolve and it is safe to return
        if (ctx == null) {
            return;
        }

        if (ctx.native_datatype_element() != null) {
            PlSqlParser.Precision_partContext precisionPart = ctx.precision_part();
            if (ctx.native_datatype_element().INT() != null
                    || ctx.native_datatype_element().INTEGER() != null
                    || ctx.native_datatype_element().SMALLINT() != null
                    || ctx.native_datatype_element().NUMERIC() != null
                    || ctx.native_datatype_element().DECIMAL() != null) {
                // NUMERIC and DECIMAL types have by default zero scale
                columnEditor.jdbcType(Types.NUMERIC).type("NUMBER");

                if (precisionPart != null) {
                    setPrecision(precisionPart, columnEditor);
                    setScale(precisionPart, columnEditor);
                }
            } else if (ctx.native_datatype_element().DATE() != null) {
                // JDBC driver reports type as timestamp but name DATE
                columnEditor.jdbcType(Types.TIMESTAMP).type("DATE");
            } else if (ctx.native_datatype_element().TIMESTAMP() != null) {
                if (ctx.WITH() != null && ctx.TIME() != null && ctx.ZONE() != null) {
                    if (ctx.LOCAL() != null) {
                        columnEditor
                                .jdbcType(OracleTypes.TIMESTAMPLTZ)
                                .type("TIMESTAMP WITH LOCAL TIME ZONE");
                    } else {
                        columnEditor
                                .jdbcType(OracleTypes.TIMESTAMPTZ)
                                .type("TIMESTAMP WITH TIME ZONE");
                    }
                } else {
                    columnEditor.jdbcType(Types.TIMESTAMP).type("TIMESTAMP");
                }

                if (precisionPart == null) {
                    columnEditor.length(6);
                } else {
                    setPrecision(precisionPart, columnEditor);
                }
            }
            // VARCHAR is the same as VARCHAR2 in Oracle
            else if (ctx.native_datatype_element().VARCHAR2() != null
                    || ctx.native_datatype_element().VARCHAR() != null) {
                columnEditor.jdbcType(Types.VARCHAR).type("VARCHAR2");

                if (precisionPart == null) {
                    columnEditor.length(getVarCharDefaultLength());
                } else {
                    setPrecision(precisionPart, columnEditor);
                }
            } else if (ctx.native_datatype_element().NVARCHAR2() != null) {
                columnEditor.jdbcType(Types.NVARCHAR).type("NVARCHAR2");

                if (precisionPart == null) {
                    columnEditor.length(getVarCharDefaultLength());
                } else {
                    setPrecision(precisionPart, columnEditor);
                }
            } else if (ctx.native_datatype_element().CHAR() != null) {
                columnEditor.jdbcType(Types.CHAR).type("CHAR").length(1);

                if (precisionPart != null) {
                    setPrecision(precisionPart, columnEditor);
                }
            } else if (ctx.native_datatype_element().NCHAR() != null) {
                columnEditor.jdbcType(Types.NCHAR).type("NCHAR").length(1);

                if (precisionPart != null) {
                    setPrecision(precisionPart, columnEditor);
                }
            } else if (ctx.native_datatype_element().BINARY_FLOAT() != null) {
                columnEditor.jdbcType(OracleTypes.BINARY_FLOAT).type("BINARY_FLOAT");
            } else if (ctx.native_datatype_element().BINARY_DOUBLE() != null) {
                columnEditor.jdbcType(OracleTypes.BINARY_DOUBLE).type("BINARY_DOUBLE");
            }
            // PRECISION keyword is mandatory
            else if (ctx.native_datatype_element().FLOAT() != null
                    || (ctx.native_datatype_element().DOUBLE() != null
                            && ctx.native_datatype_element().PRECISION() != null)) {
                columnEditor.jdbcType(Types.FLOAT).type("FLOAT");

                // TODO float's precision is about bits not decimal digits; should be ok for now to
                // over-size
                if (precisionPart != null) {
                    setPrecision(precisionPart, columnEditor);
                }
            } else if (ctx.native_datatype_element().REAL() != null) {
                columnEditor
                        .jdbcType(Types.FLOAT)
                        .type("FLOAT")
                        // TODO float's precision is about bits not decimal digits; should be ok for
                        // now to over-size
                        .length(63);
            } else if (ctx.native_datatype_element().NUMBER() != null) {
                columnEditor.jdbcType(Types.NUMERIC).type("NUMBER");

                if (precisionPart != null) {
                    if (precisionPart.ASTERISK() != null) {
                        // when asterisk is used, explicitly set precision to 38
                        columnEditor.length(38);
                    } else {
                        setPrecision(precisionPart, columnEditor);
                    }
                    setScale(precisionPart, columnEditor);
                }
            } else if (ctx.native_datatype_element().BLOB() != null) {
                columnEditor.jdbcType(Types.BLOB).type("BLOB");
            } else if (ctx.native_datatype_element().CLOB() != null) {
                columnEditor.jdbcType(Types.CLOB).type("CLOB");
            } else if (ctx.native_datatype_element().NCLOB() != null) {
                columnEditor.jdbcType(Types.NCLOB).type("NCLOB");
            } else if (ctx.native_datatype_element().RAW() != null) {
                columnEditor.jdbcType(OracleTypes.RAW).type("RAW");

                setPrecision(precisionPart, columnEditor);
            } else if (ctx.native_datatype_element().SDO_GEOMETRY() != null) {
                // Allows the registration of new SDO_GEOMETRY columns via an CREATE/ALTER TABLE
                // This is the same registration of the column that is resolved during JDBC metadata
                // inspection.
                columnEditor.jdbcType(OracleTypes.OTHER).type("SDO_GEOMETRY").length(1);
            } else if (ctx.native_datatype_element().ROWID() != null) {
                columnEditor.jdbcType(Types.VARCHAR).type("ROWID");
            } else {
                columnEditor
                        .jdbcType(OracleTypes.OTHER)
                        .type(ctx.native_datatype_element().getText());
            }
        } else if (ctx.INTERVAL() != null
                && ctx.YEAR() != null
                && ctx.TO() != null
                && ctx.MONTH() != null) {
            columnEditor.jdbcType(OracleTypes.INTERVALYM).type("INTERVAL YEAR TO MONTH").length(2);
            if (!ctx.expression().isEmpty()) {
                columnEditor.length(Integer.valueOf((ctx.expression(0).getText())));
            }
        } else if (ctx.INTERVAL() != null
                && ctx.DAY() != null
                && ctx.TO() != null
                && ctx.SECOND() != null) {
            columnEditor
                    .jdbcType(OracleTypes.INTERVALDS)
                    .type("INTERVAL DAY TO SECOND")
                    .length(2)
                    .scale(6);
            for (final PlSqlParser.ExpressionContext e : ctx.expression()) {
                if (e.getSourceInterval().startsAfter(ctx.TO().getSourceInterval())) {
                    columnEditor.scale(Integer.valueOf(e.getText()));
                } else {
                    columnEditor.length(Integer.valueOf(e.getText()));
                }
            }
            if (!ctx.expression().isEmpty()) {
                columnEditor.length(Integer.valueOf((ctx.expression(0).getText())));
            }
        } else {
            columnEditor.jdbcType(OracleTypes.OTHER).type(ctx.getText());
        }
    }

    public Column getColumn() {
        return columnEditor.create();
    }

    private int getVarCharDefaultLength() {
        // TODO replace with value from select name, value from v$parameter where
        // name='max_string_size';
        return 4000;
    }

    private void setPrecision(
            PlSqlParser.Precision_partContext precisionPart, ColumnEditor columnEditor) {
        columnEditor.length(Integer.valueOf(precisionPart.numeric(0).getText()));
    }

    private void setScale(
            PlSqlParser.Precision_partContext precisionPart, ColumnEditor columnEditor) {
        if (precisionPart.numeric().size() > 1) {
            columnEditor.scale(Integer.valueOf(precisionPart.numeric(1).getText()));
        } else if (precisionPart.numeric_negative() != null) {
            columnEditor.scale(Integer.valueOf(precisionPart.numeric_negative().getText()));
        } else {
            columnEditor.scale(0);
        }
    }
}
